@@ -1,10 +1,40 @@
+"""
+Ported to Python 3.
+"""
 
-import simplejson
+from __future__ import absolute_import
+from __future__ import division
+from __future__ import print_function
+from __future__ import unicode_literals
+
+from future.utils import PY2
+if PY2:
+    from future.builtins import filter, map, zip, ascii, chr, hex, input, next, oct, open, pow, round, super, bytes, dict, list, object, range, str, max, min  # noqa: F401
+
+
+import json
 import os.path, shutil
+
+from bs4 import BeautifulSoup
+
 from twisted.trial import unittest
 from twisted.internet import defer
+
+from zope.interface import implementer
+from twisted.web.resource import (
+    Resource,
+)
+from twisted.web.template import (
+    renderElement,
+)
+
 from allmydata import check_results, uri
 from allmydata import uri as tahoe_uri
+from allmydata.interfaces import (
+    IServer,
+    ICheckResults,
+    ICheckAndRepairResults,
+)
 from allmydata.util import base32
 from allmydata.web import check_results as web_check_results
 from allmydata.storage_client import StorageFarmBroker, NativeStorageServer
@@ -12,79 +42,173 @@ from allmydata.storage.server import storage_index_to_dir
 from allmydata.monitor import Monitor
 from allmydata.test.no_network import GridTestMixin
 from allmydata.immutable.upload import Data
-from allmydata.test.common_web import WebRenderingMixin
 from allmydata.mutable.publish import MutableData
 
-class FakeClient:
+from .common import (
+    EMPTY_CLIENT_CONFIG,
+)
+from .common_web import (
+    render,
+)
+
+from .web.common import (
+    assert_soup_has_favicon,
+    assert_soup_has_tag_with_content,
+)
+
+class FakeClient(object):
     def get_storage_broker(self):
         return self.storage_broker
 
-class WebResultsRendering(unittest.TestCase, WebRenderingMixin):
+
+@implementer(IServer)
+class FakeServer(object):  # type: ignore # incomplete implementation
+
+    def get_name(self):
+        return "fake name"
+
+    def get_longname(self):
+        return "fake longname"
+
+    def get_nickname(self):
+        return "fake nickname"
+
+
+@implementer(ICheckResults)
+class FakeCheckResults(object):  # type: ignore # incomplete implementation
+
+    def __init__(self, si=None,
+                 healthy=False, recoverable=False,
+                 summary="fake summary"):
+        self._storage_index = si
+        self._is_healthy = healthy
+        self._is_recoverable = recoverable
+        self._summary = summary
+
+    def get_storage_index(self):
+        return self._storage_index
+
+    def get_storage_index_string(self):
+        return base32.b2a_or_none(self._storage_index)
+
+    def is_healthy(self):
+        return self._is_healthy
+
+    def is_recoverable(self):
+        return self._is_recoverable
+
+    def get_summary(self):
+        return self._summary
+
+    def get_corrupt_shares(self):
+        # returns a list of (IServer, storage_index, sharenum)
+        return [(FakeServer(), b"<fake-si>", 0)]
+
+
+@implementer(ICheckAndRepairResults)
+class FakeCheckAndRepairResults(object):  # type: ignore # incomplete implementation
+
+    def __init__(self, si=None,
+                 repair_attempted=False,
+                 repair_success=False):
+        self._storage_index = si
+        self._repair_attempted = repair_attempted
+        self._repair_success = repair_success
+
+    def get_storage_index(self):
+        return self._storage_index
+
+    def get_pre_repair_results(self):
+        return FakeCheckResults()
+
+    def get_post_repair_results(self):
+        return FakeCheckResults()
+
+    def get_repair_attempted(self):
+        return self._repair_attempted
+
+    def get_repair_successful(self):
+        return self._repair_success
+
+
+class ElementResource(Resource, object):
+    def __init__(self, element):
+        Resource.__init__(self)
+        self.element = element
+
+    def render(self, request):
+        return renderElement(request, self.element)
+
+
+class WebResultsRendering(unittest.TestCase):
+
+    @staticmethod
+    def remove_tags(html):
+        return BeautifulSoup(html, 'html5lib').get_text(separator=" ")
 
     def create_fake_client(self):
-        sb = StorageFarmBroker(None, True)
+        sb = StorageFarmBroker(True, None, EMPTY_CLIENT_CONFIG)
         # s.get_name() (the "short description") will be "v0-00000000".
         # s.get_longname() will include the -long suffix.
-        # s.get_peerid() (i.e. tubid) will be "aaa.." or "777.." or "ceir.."
-        servers = [("v0-00000000-long", "\x00"*20, "peer-0"),
-                   ("v0-ffffffff-long", "\xff"*20, "peer-f"),
-                   ("v0-11111111-long", "\x11"*20, "peer-11")]
-        for (key_s, peerid, nickname) in servers:
-            tubid_b32 = base32.b2a(peerid)
-            furl = "pb://%s@nowhere/fake" % tubid_b32
+        servers = [(b"v0-00000000-long", b"\x00"*20, "peer-0"),
+                   (b"v0-ffffffff-long", b"\xff"*20, "peer-f"),
+                   (b"v0-11111111-long", b"\x11"*20, "peer-11")]
+        for (key_s, binary_tubid, nickname) in servers:
+            server_id = key_s
+            tubid_b32 = base32.b2a(binary_tubid)
+            furl = "pb://%s@nowhere/fake" % str(tubid_b32, "utf-8")
             ann = { "version": 0,
                     "service-name": "storage",
                     "anonymous-storage-FURL": furl,
                     "permutation-seed-base32": "",
-                    "nickname": unicode(nickname),
+                    "nickname": str(nickname),
                     "app-versions": {}, # need #466 and v2 introducer
                     "my-version": "ver",
                     "oldest-supported": "oldest",
                     }
-            s = NativeStorageServer(key_s, ann)
-            sb.test_add_server(peerid, s) # XXX: maybe use key_s?
+            s = NativeStorageServer(server_id, ann, None, None, None)
+            sb.test_add_server(server_id, s)
         c = FakeClient()
         c.storage_broker = sb
         return c
 
-    def render_json(self, page):
-        d = self.render1(page, args={"output": ["json"]})
-        return d
+    def render_json(self, resource):
+        return self.successResultOf(render(resource, {b"output": [b"json"]}))
+
+    def render_element(self, element, args=None):
+        if args is None:
+            args = {}
+        return self.successResultOf(render(ElementResource(element), args))
 
     def test_literal(self):
+        lcr = web_check_results.LiteralCheckResultsRendererElement()
+
+        html = self.render_element(lcr)
+        self.failUnlessIn(b"Literal files are always healthy", html)
+
+        html = self.render_element(lcr, args={b"return_to": [b"FOOURL"]})
+        self.failUnlessIn(b"Literal files are always healthy", html)
+        self.failUnlessIn(b'<a href="FOOURL">Return to file.</a>', html)
+
         c = self.create_fake_client()
         lcr = web_check_results.LiteralCheckResultsRenderer(c)
 
-        d = self.render1(lcr)
-        def _check(html):
-            s = self.remove_tags(html)
-            self.failUnlessIn("Literal files are always healthy", s)
-        d.addCallback(_check)
-        d.addCallback(lambda ignored:
-                      self.render1(lcr, args={"return_to": ["FOOURL"]}))
-        def _check_return_to(html):
-            s = self.remove_tags(html)
-            self.failUnlessIn("Literal files are always healthy", s)
-            self.failUnlessIn('<a href="FOOURL">Return to file.</a>',
-                              html)
-        d.addCallback(_check_return_to)
-        d.addCallback(lambda ignored: self.render_json(lcr))
-        def _check_json(json):
-            j = simplejson.loads(json)
-            self.failUnlessEqual(j["storage-index"], "")
-            self.failUnlessEqual(j["results"]["healthy"], True)
-        d.addCallback(_check_json)
-        return d
+        js = self.render_json(lcr)
+        j = json.loads(js)
+        self.failUnlessEqual(j["storage-index"], "")
+        self.failUnlessEqual(j["results"]["healthy"], True)
+
 
     def test_check(self):
         c = self.create_fake_client()
         sb = c.storage_broker
-        serverid_1 = "\x00"*20
-        serverid_f = "\xff"*20
+        serverid_1 = b"\x00"*20
+        serverid_f = b"\xff"*20
         server_1 = sb.get_stub_server(serverid_1)
         server_f = sb.get_stub_server(serverid_f)
-        u = uri.CHKFileURI("\x00"*16, "\x00"*32, 3, 10, 1234)
-        data = { "count_shares_needed": 3,
+        u = uri.CHKFileURI(b"\x00"*16, b"\x00"*32, 3, 10, 1234)
+        data = { "count_happiness": 8,
+                 "count_shares_needed": 3,
                  "count_shares_expected": 9,
                  "count_shares_good": 10,
                  "count_good_share_hosts": 11,
@@ -101,29 +225,28 @@ class WebResultsRendering(unittest.TestCase, WebRenderingMixin):
                  }
         cr = check_results.CheckResults(u, u.get_storage_index(),
                                         healthy=True, recoverable=True,
-                                        needs_rebalancing=False,
                                         summary="groovy",
                                         **data)
-        w = web_check_results.CheckResultsRenderer(c, cr)
-        html = self.render2(w)
+        w = web_check_results.CheckResultsRendererElement(c, cr)
+        html = self.render_element(w)
         s = self.remove_tags(html)
         self.failUnlessIn("File Check Results for SI=2k6avp", s) # abbreviated
         self.failUnlessIn("Healthy : groovy", s)
         self.failUnlessIn("Share Counts: need 3-of-9, have 10", s)
+        self.failUnlessIn("Happiness Level: 8", s)
         self.failUnlessIn("Hosts with good shares: 11", s)
         self.failUnlessIn("Corrupt shares: none", s)
         self.failUnlessIn("Wrong Shares: 0", s)
         self.failUnlessIn("Recoverable Versions: 1", s)
         self.failUnlessIn("Unrecoverable Versions: 0", s)
-        self.failUnlessIn("Good Shares (sorted in share order): Share ID Nickname Node ID shareid1 peer-0 00000000 peer-f ffffffff", s)
+        self.failUnlessIn("Good Shares (sorted in share order):  Share ID Nickname Node ID shareid1 peer-0 00000000 peer-f ffffffff", s)
 
         cr = check_results.CheckResults(u, u.get_storage_index(),
                                         healthy=False, recoverable=True,
-                                        needs_rebalancing=False,
                                         summary="ungroovy",
                                         **data)
-        w = web_check_results.CheckResultsRenderer(c, cr)
-        html = self.render2(w)
+        w = web_check_results.CheckResultsRendererElement(c, cr)
+        html = self.render_element(w)
         s = self.remove_tags(html)
         self.failUnlessIn("File Check Results for SI=2k6avp", s) # abbreviated
         self.failUnlessIn("Not Healthy! : ungroovy", s)
@@ -132,32 +255,32 @@ class WebResultsRendering(unittest.TestCase, WebRenderingMixin):
         data["list_corrupt_shares"] = [(server_1, u.get_storage_index(), 2)]
         cr = check_results.CheckResults(u, u.get_storage_index(),
                                         healthy=False, recoverable=False,
-                                        needs_rebalancing=False,
                                         summary="rather dead",
                                         **data)
-        w = web_check_results.CheckResultsRenderer(c, cr)
-        html = self.render2(w)
+        w = web_check_results.CheckResultsRendererElement(c, cr)
+        html = self.render_element(w)
         s = self.remove_tags(html)
         self.failUnlessIn("File Check Results for SI=2k6avp", s) # abbreviated
         self.failUnlessIn("Not Recoverable! : rather dead", s)
-        self.failUnlessIn("Corrupt shares: Share ID Nickname Node ID sh#2 peer-0 00000000", s)
+        self.failUnlessIn("Corrupt shares:  Share ID Nickname Node ID sh#2 peer-0 00000000", s)
 
-        html = self.render2(w)
+        html = self.render_element(w)
         s = self.remove_tags(html)
         self.failUnlessIn("File Check Results for SI=2k6avp", s) # abbreviated
         self.failUnlessIn("Not Recoverable! : rather dead", s)
 
-        html = self.render2(w, args={"return_to": ["FOOURL"]})
-        self.failUnlessIn('<a href="FOOURL">Return to file/directory.</a>',
+        html = self.render_element(w, args={b"return_to": [b"FOOURL"]})
+        self.failUnlessIn(b'<a href="FOOURL">Return to file/directory.</a>',
                           html)
 
+        w = web_check_results.CheckResultsRenderer(c, cr)
         d = self.render_json(w)
         def _check_json(jdata):
-            j = simplejson.loads(jdata)
+            j = json.loads(jdata)
             self.failUnlessEqual(j["summary"], "rather dead")
             self.failUnlessEqual(j["storage-index"],
                                  "2k6avpjga3dho3zsjo6nnkt7n4")
-            expected = {'needs-rebalancing': False,
+            expected = {'count-happiness': 8,
                         'count-shares-expected': 9,
                         'healthy': False,
                         'count-unrecoverable-versions': 0,
@@ -175,24 +298,25 @@ class WebResultsRendering(unittest.TestCase, WebRenderingMixin):
                         'recoverable': False,
                         }
             self.failUnlessEqual(j["results"], expected)
-        d.addCallback(_check_json)
-        d.addCallback(lambda ignored: self.render1(w))
+        _check_json(d)
+
+        w = web_check_results.CheckResultsRendererElement(c, cr)
+        d = self.render_element(w)
         def _check(html):
             s = self.remove_tags(html)
             self.failUnlessIn("File Check Results for SI=2k6avp", s)
             self.failUnlessIn("Not Recoverable! : rather dead", s)
-        d.addCallback(_check)
-        return d
-
+        _check(html)
 
     def test_check_and_repair(self):
         c = self.create_fake_client()
         sb = c.storage_broker
-        serverid_1 = "\x00"*20
-        serverid_f = "\xff"*20
-        u = uri.CHKFileURI("\x00"*16, "\x00"*32, 3, 10, 1234)
+        serverid_1 = b"\x00"*20
+        serverid_f = b"\xff"*20
+        u = uri.CHKFileURI(b"\x00"*16, b"\x00"*32, 3, 10, 1234)
 
-        data = { "count_shares_needed": 3,
+        data = { "count_happiness": 5,
+                 "count_shares_needed": 3,
                  "count_shares_expected": 10,
                  "count_shares_good": 6,
                  "count_good_share_hosts": 7,
@@ -210,11 +334,11 @@ class WebResultsRendering(unittest.TestCase, WebRenderingMixin):
                  }
         pre_cr = check_results.CheckResults(u, u.get_storage_index(),
                                             healthy=False, recoverable=True,
-                                            needs_rebalancing=False,
                                             summary="illing",
                                             **data)
 
-        data = { "count_shares_needed": 3,
+        data = { "count_happiness": 9,
+                 "count_shares_needed": 3,
                  "count_shares_expected": 10,
                  "count_shares_good": 10,
                  "count_good_share_hosts": 11,
@@ -232,7 +356,6 @@ class WebResultsRendering(unittest.TestCase, WebRenderingMixin):
                  }
         post_cr = check_results.CheckResults(u, u.get_storage_index(),
                                              healthy=True, recoverable=True,
-                                             needs_rebalancing=False,
                                              summary="groovy",
                                              **data)
 
@@ -241,8 +364,8 @@ class WebResultsRendering(unittest.TestCase, WebRenderingMixin):
         crr.post_repair_results = post_cr
         crr.repair_attempted = False
 
-        w = web_check_results.CheckAndRepairResultsRenderer(c, crr)
-        html = self.render2(w)
+        w = web_check_results.CheckAndRepairResultsRendererElement(c, crr)
+        html = self.render_element(w)
         s = self.remove_tags(html)
 
         self.failUnlessIn("File Check-And-Repair Results for SI=2k6avp", s)
@@ -253,7 +376,7 @@ class WebResultsRendering(unittest.TestCase, WebRenderingMixin):
 
         crr.repair_attempted = True
         crr.repair_successful = True
-        html = self.render2(w)
+        html = self.render_element(w)
         s = self.remove_tags(html)
 
         self.failUnlessIn("File Check-And-Repair Results for SI=2k6avp", s)
@@ -265,11 +388,10 @@ class WebResultsRendering(unittest.TestCase, WebRenderingMixin):
         crr.repair_successful = False
         post_cr = check_results.CheckResults(u, u.get_storage_index(),
                                              healthy=False, recoverable=True,
-                                             needs_rebalancing=False,
                                              summary="better",
                                              **data)
         crr.post_repair_results = post_cr
-        html = self.render2(w)
+        html = self.render_element(w)
         s = self.remove_tags(html)
 
         self.failUnlessIn("File Check-And-Repair Results for SI=2k6avp", s)
@@ -281,11 +403,10 @@ class WebResultsRendering(unittest.TestCase, WebRenderingMixin):
         crr.repair_successful = False
         post_cr = check_results.CheckResults(u, u.get_storage_index(),
                                              healthy=False, recoverable=False,
-                                             needs_rebalancing=False,
                                              summary="worse",
                                              **data)
         crr.post_repair_results = post_cr
-        html = self.render2(w)
+        html = self.render_element(w)
         s = self.remove_tags(html)
 
         self.failUnlessIn("File Check-And-Repair Results for SI=2k6avp", s)
@@ -293,24 +414,218 @@ class WebResultsRendering(unittest.TestCase, WebRenderingMixin):
         self.failUnlessIn("Repair unsuccessful", s)
         self.failUnlessIn("Post-Repair Checker Results:", s)
 
-        d = self.render_json(w)
-        def _got_json(data):
-            j = simplejson.loads(data)
-            self.failUnlessEqual(j["repair-attempted"], True)
-            self.failUnlessEqual(j["storage-index"],
-                                 "2k6avpjga3dho3zsjo6nnkt7n4")
-            self.failUnlessEqual(j["pre-repair-results"]["summary"], "illing")
-            self.failUnlessEqual(j["post-repair-results"]["summary"], "worse")
-        d.addCallback(_got_json)
+        w = web_check_results.CheckAndRepairResultsRenderer(c, crr)
+        j = json.loads(self.render_json(w))
+        self.failUnlessEqual(j["repair-attempted"], True)
+        self.failUnlessEqual(j["storage-index"],
+                             "2k6avpjga3dho3zsjo6nnkt7n4")
+        self.failUnlessEqual(j["pre-repair-results"]["summary"], "illing")
+        self.failUnlessEqual(j["post-repair-results"]["summary"], "worse")
+
+        w = web_check_results.CheckAndRepairResultsRenderer(c, None)
+        j = json.loads(self.render_json(w))
+        self.failUnlessEqual(j["repair-attempted"], False)
+        self.failUnlessEqual(j["storage-index"], "")
+
+
+    def test_deep_check_renderer(self):
+        status = check_results.DeepCheckResults(b"fake-root-si")
+        status.add_check(
+            FakeCheckResults(b"<unhealthy/unrecoverable>", False, False),
+            (u"fake", u"unhealthy", u"unrecoverable")
+        )
+        status.add_check(
+            FakeCheckResults(b"<healthy/recoverable>", True, True),
+            (u"fake", u"healthy", u"recoverable")
+        )
+        status.add_check(
+            FakeCheckResults(b"<healthy/unrecoverable>", True, False),
+            (u"fake", u"healthy", u"unrecoverable")
+        )
+        status.add_check(
+            FakeCheckResults(b"<unhealthy/unrecoverable>", False, True),
+            (u"fake", u"unhealthy", u"recoverable")
+        )
+
+        monitor = Monitor()
+        monitor.set_status(status)
+
+        elem = web_check_results.DeepCheckResultsRendererElement(monitor)
+        doc = self.render_element(elem)
+        soup = BeautifulSoup(doc, 'html5lib')
+
+        assert_soup_has_favicon(self, soup)
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"title",
+            u"Tahoe-LAFS - Deep Check Results"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"h1",
+            "Deep-Check Results for root SI="
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"Objects Checked: 4"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"Objects Healthy: 2"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"Objects Unhealthy: 2"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"Objects Unrecoverable: 2"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"Corrupt Shares: 4"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"h2",
+            u"Files/Directories That Had Problems:"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"fake/unhealthy/recoverable: fake summary"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"fake/unhealthy/unrecoverable: fake summary"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"h2",
+            u"Servers on which corrupt shares were found"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"h2",
+            u"Corrupt Shares"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"h2",
+            u"All Results"
+        )
+
+    def test_deep_check_and_repair_renderer(self):
+        status = check_results.DeepCheckAndRepairResults(b"")
+
+        status.add_check_and_repair(
+            FakeCheckAndRepairResults(b"attempted/success", True, True),
+            (u"attempted", u"success")
+        )
+        status.add_check_and_repair(
+            FakeCheckAndRepairResults(b"attempted/failure", True, False),
+            (u"attempted", u"failure")
+        )
+        status.add_check_and_repair(
+            FakeCheckAndRepairResults(b"unattempted/failure", False, False),
+            (u"unattempted", u"failure")
+        )
+
+        monitor = Monitor()
+        monitor.set_status(status)
+
+        elem = web_check_results.DeepCheckAndRepairResultsRendererElement(monitor)
+        doc = self.render_element(elem)
+        soup = BeautifulSoup(doc, 'html5lib')
+
+        assert_soup_has_favicon(self, soup)
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"title",
+            u"Tahoe-LAFS - Deep Check Results"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"h1",
+            u"Deep-Check-And-Repair Results for root SI="
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"Objects Checked: 3"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"Objects Healthy (before repair): 0"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"Objects Unhealthy (before repair): 3"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"Corrupt Shares (before repair): 3"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"Repairs Attempted: 2"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"Repairs Successful: 1"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            "Repairs Unsuccessful: 1"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"Objects Healthy (after repair): 0"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"Objects Unhealthy (after repair): 3"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"li",
+            u"Corrupt Shares (after repair): 3"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"h2",
+            u"Files/Directories That Had Problems:"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"h2",
+            u"Files/Directories That Still Have Problems:"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"h2",
+            u"Servers on which corrupt shares were found"
+        )
+
+        assert_soup_has_tag_with_content(
+            self, soup, u"h2",
+            u"Remaining Corrupt Shares"
+        )
 
-        w2 = web_check_results.CheckAndRepairResultsRenderer(c, None)
-        d.addCallback(lambda ignored: self.render_json(w2))
-        def _got_lit_results(data):
-            j = simplejson.loads(data)
-            self.failUnlessEqual(j["repair-attempted"], False)
-            self.failUnlessEqual(j["storage-index"], "")
-        d.addCallback(_got_lit_results)
-        return d
 
 class BalancingAct(GridTestMixin, unittest.TestCase):
     # test for #1115 regarding the 'count-good-share-hosts' metric
@@ -357,7 +672,7 @@ class BalancingAct(GridTestMixin, unittest.TestCase):
             "This little printing function is only meant for < 26 servers"
         shares_chart = {}
         names = dict(zip([ss.my_nodeid
-                          for _,ss in self.g.servers_by_number.iteritems()],
+                          for _,ss in self.g.servers_by_number.items()],
                          letters))
         for shnum, serverid, _ in self.find_uri_shares(uri):
             shares_chart.setdefault(shnum, []).append(names[serverid])
@@ -367,12 +682,12 @@ class BalancingAct(GridTestMixin, unittest.TestCase):
         self.basedir = "checker/BalancingAct/1115"
         self.set_up_grid(num_servers=1)
         c0 = self.g.clients[0]
-        c0.DEFAULT_ENCODING_PARAMETERS['happy'] = 1
-        c0.DEFAULT_ENCODING_PARAMETERS['n'] = 4
-        c0.DEFAULT_ENCODING_PARAMETERS['k'] = 3
+        c0.encoding_params['happy'] = 1
+        c0.encoding_params['n'] = 4
+        c0.encoding_params['k'] = 3
 
-        DATA = "data" * 100
-        d = c0.upload(Data(DATA, convergence=""))
+        DATA = b"data" * 100
+        d = c0.upload(Data(DATA, convergence=b""))
         def _stash_immutable(ur):
             self.imm = c0.create_node_from_uri(ur.get_uri())
             self.uri = self.imm.get_uri()
@@ -386,7 +701,7 @@ class BalancingAct(GridTestMixin, unittest.TestCase):
         def add_three(_, i):
             # Add a new server with just share 3
             self.add_server_with_share(i, self.uri, 3)
-            #print self._pretty_shares_chart(self.uri)
+            #print(self._pretty_shares_chart(self.uri))
         for i in range(1,5):
             d.addCallback(add_three, i)
 
@@ -394,7 +709,6 @@ class BalancingAct(GridTestMixin, unittest.TestCase):
             return self.imm.check_and_repair(Monitor())
         def _check_counts(crr, shares_good, good_share_hosts):
             prr = crr.get_post_repair_results()
-            #print self._pretty_shares_chart(self.uri)
             self.failUnlessEqual(prr.get_share_counter_good(), shares_good)
             self.failUnlessEqual(prr.get_host_counter_good_shares(),
                                  good_share_hosts)
@@ -404,16 +718,26 @@ class BalancingAct(GridTestMixin, unittest.TestCase):
             0:[A] 1:[A] 2:[A] 3:[A,B,C,D,E]
           4 good shares, but 5 good hosts
         After deleting all instances of share #3 and repairing:
-            0:[A,B], 1:[A,C], 2:[A,D], 3:[E]
-          Still 4 good shares and 5 good hosts
+            0:[A], 1:[A,B], 2:[C,A], 3:[E]
+# actually: {0: ['E', 'A'], 1: ['C', 'A'], 2: ['A', 'B'], 3: ['D']}
+          Still 4 good shares but now 4 good hosts
             """
         d.addCallback(_check_and_repair)
         d.addCallback(_check_counts, 4, 5)
         d.addCallback(lambda _: self.delete_shares_numbered(self.uri, [3]))
         d.addCallback(_check_and_repair)
-        d.addCallback(_check_counts, 4, 5)
-        d.addCallback(lambda _: [self.g.break_server(sid)
-                                 for sid in self.g.get_all_serverids()])
+
+        # it can happen that our uploader will choose, e.g., to upload
+        # to servers B, C, D, E .. which will mean that all 5 serves
+        # now contain our shares (and thus "respond").
+
+        def _check_happy(crr):
+            prr = crr.get_post_repair_results()
+            self.assertTrue(prr.get_host_counter_good_shares() >= 4)
+            return crr
+        d.addCallback(_check_happy)
+        d.addCallback(lambda _: all([self.g.break_server(sid)
+                                     for sid in self.g.get_all_serverids()]))
         d.addCallback(_check_and_repair)
         d.addCallback(_check_counts, 0, 0)
         return d
@@ -426,15 +750,15 @@ class AddLease(GridTestMixin, unittest.TestCase):
         self.basedir = "checker/AddLease/875"
         self.set_up_grid(num_servers=1)
         c0 = self.g.clients[0]
-        c0.DEFAULT_ENCODING_PARAMETERS['happy'] = 1
+        c0.encoding_params['happy'] = 1
         self.uris = {}
-        DATA = "data" * 100
-        d = c0.upload(Data(DATA, convergence=""))
+        DATA = b"data" * 100
+        d = c0.upload(Data(DATA, convergence=b""))
         def _stash_immutable(ur):
             self.imm = c0.create_node_from_uri(ur.get_uri())
         d.addCallback(_stash_immutable)
         d.addCallback(lambda ign:
-            c0.create_mutable_file(MutableData("contents")))
+            c0.create_mutable_file(MutableData(b"contents")))
         def _stash_mutable(node):
             self.mut = node
         d.addCallback(_stash_mutable)
@@ -449,13 +773,13 @@ class AddLease(GridTestMixin, unittest.TestCase):
         d.addCallback(_check_cr, "mutable-normal")
 
         really_did_break = []
-        # now break the server's remote_add_lease call
+        # now break the server's add_lease call
         def _break_add_lease(ign):
             def broken_add_lease(*args, **kwargs):
                 really_did_break.append(1)
                 raise KeyError("intentional failure, should be ignored")
-            assert self.g.servers_by_number[0].remote_add_lease
-            self.g.servers_by_number[0].remote_add_lease = broken_add_lease
+            assert self.g.servers_by_number[0].add_lease
+            self.g.servers_by_number[0].add_lease = broken_add_lease
         d.addCallback(_break_add_lease)
 
         # and confirm that the files still look healthy
@@ -514,14 +838,14 @@ class TooParallel(GridTestMixin, unittest.TestCase):
         def _start(ign):
             self.set_up_grid(num_servers=4)
             self.c0 = self.g.clients[0]
-            self.c0.DEFAULT_ENCODING_PARAMETERS = { "k": 1,
-                                               "happy": 4,
-                                               "n": 4,
-                                               "max_segment_size": 5,
-                                               }
+            self.c0.encoding_params = { "k": 1,
+                                        "happy": 4,
+                                        "n": 4,
+                                        "max_segment_size": 5,
+                                      }
             self.uris = {}
-            DATA = "data" * 100 # 400/5 = 80 blocks
-            return self.c0.upload(Data(DATA, convergence=""))
+            DATA = b"data" * 100 # 400/5 = 80 blocks
+            return self.c0.upload(Data(DATA, convergence=b""))
         d.addCallback(_start)
         def _do_check(ur):
             n = self.c0.create_node_from_uri(ur.get_uri())
@@ -538,5 +862,3 @@ class TooParallel(GridTestMixin, unittest.TestCase):
             return res
         d.addBoth(_clean_up)
         return d
-
-    test_immutable.timeout = 80
